# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerics for fp8."""

from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp


@utils.flax_slots_kw_only_dataclass
class FpNumericsConfig:
  nexp: int
  minexp: int
  nmant: int
  has_subnormals: bool
  has_two_nan: bool
  has_naninf: bool

e3m0 = FpNumericsConfig(
    nexp=3,
    minexp=0,
    nmant=0,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
)

e2m1 = FpNumericsConfig(
    nexp=2,
    minexp=0,
    nmant=1,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
)

e1m0 = FpNumericsConfig(
    nexp=1,
    minexp=0,
    nmant=0,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
)

float8_e4m3fn = FpNumericsConfig(
    nexp=4,
    minexp=-6,
    nmant=3,
    has_subnormals=True,
    has_two_nan=True,
    has_naninf=False,
)

float8_e5m2 = FpNumericsConfig(
    nexp=5,
    minexp=-14,
    nmant=2,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=True,
)

float16 = FpNumericsConfig(
    nexp=5,
    minexp=-14,
    nmant=10,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=True,
)


# TODO(lew): Add minexp.
def fp_round(
    x,
    *,
    cfg: FpNumericsConfig,
    key: jax.Array,
    stochastic_rounding: bool,
):
  """FP stochastic rounding for a given mantissa and exponent. Returns bf16."""
  nexp = cfg.nexp
  nmant = cfg.nmant
  assert cfg.minexp == 0, 'minexp not implemented'
  assert not cfg.has_subnormals, 'subnormals not implemented'
  assert not cfg.has_two_nan, 'two_nan not implemented'
  assert not cfg.has_naninf, 'naninf not implemented'

  input_dtype = x.dtype
  total_bits = jnp.finfo(input_dtype).bits
  # (sign, exp, mantissa) = (1,8,x)
  msg = f'Unsupported dtype for sthochastic rounding: {input_dtype}.'
  assert input_dtype == jnp.bfloat16 or input_dtype == jnp.float32, msg
  # bf16 bit pattern: seeeeeeeemmmmmmm
  #                   7654321076543210
  container_man = jnp.finfo(input_dtype).nmant
  bits_dtype = jnp.uint16 if total_bits == 16 else jnp.uint32
  assert nmant <= container_man, 'bf16 man bits'
  man_trunc_bits = container_man - nmant
  man_trunc_mask = bits_dtype((1 << man_trunc_bits) - 1)

  # TODO(lew): We can use smaller dtype
  x = jax.lax.bitcast_convert_type(x, bits_dtype)  # adds noise to man
  # def bprint(s, x):
  #   xx = int(x)
  #   print(f'{s:20} = {xx:016b}')
  if stochastic_rounding:
    noise = jax.random.bits(key, x.shape, bits_dtype)
    noise = jax.lax.convert_element_type(noise, bits_dtype) & man_trunc_mask
  else:
    noise = 1 << (man_trunc_bits - 1)
    noise = jax.lax.convert_element_type(noise, bits_dtype)
  # This noise add might overflow up to sign bit if x(bf16) has max exp.
  # In bf16 this happens if x=nan or x=inf.
  x = x + noise
  x = jax.lax.bitwise_and(x, ~man_trunc_mask)  # zero-out lowest bits in man

  if nexp is not None:
    assert input_dtype == jnp.bfloat16
    # min_exp = jnp.uint16(0b0_10000000_0000000)
    # max_exp = jnp.uint16(0b0_10000111_0000000)
    # bf16_exp_bias = jnp.uint16(0b0_01111111)
    min_exp = jnp.uint16(0b0_01111111)
    max_exp = min_exp + 2**nexp - 1

    min_repr = min_exp << container_man
    container_mant_mask = (1 << container_man) - 1
    max_mant = jnp.uint16(container_mant_mask & ~man_trunc_mask)
    max_repr = (max_exp << container_man) + max_mant

    exp_mant_msk = jnp.uint16(0b0_11111111_1111111)
    exp = jax.lax.bitwise_and(x, exp_mant_msk)
    exp = jnp.clip(exp, min_repr, max_repr)
    x = jax.lax.bitwise_or(jax.lax.bitwise_and(x, ~exp_mant_msk), exp)
    # TODO(lew): we can do faster clipping with bit twiddling

  x = jax.lax.bitcast_convert_type(x, input_dtype)

  return x


def fp_largest_representable(cfg: FpNumericsConfig):
  nexp = cfg.nexp
  nmant = cfg.nmant
  assert cfg.minexp == 0, 'minexp not implemented'
  assert not cfg.has_subnormals, 'subnormals not implemented'
  assert not cfg.has_naninf, 'naninf not implemented'
  max_exp = 2**nexp - 1 + cfg.minexp
  max_mant = 2**nmant - 1 - (1 if cfg.has_two_nan else 0)
  return 2**max_exp * (1 + (max_mant / 2**nmant))


@utils.flax_slots_kw_only_dataclass
class FpNumerics(numerics.AqtNumerics):
  """Numerics for fp8."""

  # Requested rounding precision.
  cfg: FpNumericsConfig = utils.static_field()
  stochastic_rounding: bool = utils.static_field(default=False)
  clip_gradient: bool = utils.static_field(default=False)

  def abs_val_mapped_to(self):
    return fp_largest_representable(cfg=self.cfg)

  def get_dtype(self):
    return jnp.bfloat16

  def vjp_fwd(self, x, context):
    x = fp_round(
        x.astype(jnp.bfloat16),
        cfg=self.cfg,
        key=context.key,
        stochastic_rounding=self.stochastic_rounding,
    )
    res = (x,)
    return x, res

  def vjp_bwd(self, res, grad):
    ret = grad
    if self.clip_gradient:
      (x,) = res
      clip_bound = self.abs_val_mapped_to()
      ret *= (-clip_bound <= x) * (x <= clip_bound)
    return (ret, None)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration dataclasses."""

from typing import Literal, Sequence
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import calibration
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import int_numerics
from aqt.jax.v2.numerics import no_numerics
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp


AbstractAqtNumerics = numerics.AqtNumerics
AbstractAqtCalibration = calibration.Calibration


@utils.flax_slots_kw_only_dataclass
class Quantizer:
  """Configuration of quantization of one tensor."""

  numerics: AbstractAqtNumerics = utils.static_field()
  calib_shared_axes: Sequence[utils.AxisIdx] | Literal["per_tensor"] | None = (
      utils.static_field()
  )
  scale_stop_grad: bool = utils.static_field()
  # noise+clip+round
  # We apply gradient of clip_and_round in bwd pass.
  calibration: type[AbstractAqtCalibration] = utils.static_field()
  _calibrator: AbstractAqtCalibration | None = utils.static_field(default=None)
  # Round up the calibration to power of 2 (po2).
  po2_scale: bool = utils.static_field()
  # The dtype of the quantization scale array. If not set, the scale array will
  # be in the same dtype as the input.
  scale_dtype: jnp.dtype | None = utils.static_field(default=None)
  # TODO(yichizh): Factor out auxilliary dataclasses into a separate file.
  context: utils.Context

  # we need to speed up this initialization for the backward pass to happen
  # outside of bwd pass.
  def init_calibration(self):
    assert self._calibrator is None, "second call to self.init_calibration()"
    self._calibrator = self.calibration()

  # TODO(yichizh): Need to add type annotation back to cfg.
  def quant(
      self,
      x,
      *,
      calibration_axes,
  ) -> tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn]:
    """The core quantizing function."""
    qt = self.calibrate(x, calibration_axes=calibration_axes)
    qt, quant_grad = self.calculate_qvalue(x, qt)
    return qt, quant_grad

  def calibrate(self, x, *, calibration_axes) -> aqt_tensor.QTensor:
    """Create incomplete QTensor with only quantization parameters."""
    if isinstance(self.numerics, no_numerics.NoNumerics):
      qt = aqt_tensor.QTensor(
          qvalue=x, scale=[], scale_t=None, dequant_dtype=x.dtype
      )
      return qt

    dequant_dtype = x.dtype
    # TODO(lew): We should cast earlier. xhs_q should be in cfg.xhs.dtype
    # TODO(lew): After we implement optimization to not double-quantize,
    #   what would happen if we pass fq value (xhs_q2) in residual?
    if self.calib_shared_axes == "per_tensor":
      shared_axes = list(range(x.ndim))
    else:
      shared_axes = self.calib_shared_axes or calibration_axes

    assert self._calibrator is not None, "forgot self.init_calibration()?"
    bound = self._calibrator.get_bound(x, shared_axes, self.context)
    abs_max_mapped_to = self.numerics.abs_val_mapped_to()
    scale = bound / abs_max_mapped_to

    if self.po2_scale:
      # With floor the biggest value (we are using jnp.max) is in the range of
      # clipping and therefore have a correct gradinet.
      scale = 2 ** jnp.floor(jnp.log2(jax.lax.reciprocal(scale)))
      scale = jax.lax.reciprocal(scale)
    if self.scale_stop_grad:
      # TODO(lew): Does not matter in DG, because we are using custom gradient.
      #   We should take that into account somehow.
      scale = jax.lax.stop_gradient(scale)
    if self.scale_dtype is not None:
      scale = scale.astype(self.scale_dtype)

    qt = aqt_tensor.QTensor(
        qvalue=None,
        scale=[scale],
        scale_t=None,
        dequant_dtype=dequant_dtype,
    )
    return qt

  def calculate_qvalue(
      self,
      x,
      qt: aqt_tensor.QTensor
  ) -> tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn]:
    """Uses the quantization parameters in qt to quantize x."""
    if isinstance(self.numerics, no_numerics.NoNumerics):
      return qt, None

    # TODO: b/333984742 - make numeric as a member of QTensor, and put
    # numerics-related logics into the QTensor.
    qt = qt.quant(x)

    # TODO(lew): A logical thing would be if this call was part of
    # QTensor.quant.
    x_q, res = self.numerics.vjp_fwd(qt.qvalue, self.context)
    quant_grad = jax.tree_util.Partial(self.numerics.vjp_bwd, res)

    qt = qt.replace(qvalue=x_q)
    return qt, quant_grad


def quantizer_make(
    n_bits: int | None,
    preserve_max_val: bool = False,
    initialize_calibration: bool = True,
) -> Quantizer:
  """Makes Quantizer."""
  if n_bits is None:
    effective_numerics = no_numerics.NoNumerics()
  else:
    pz = False if n_bits == 1 else True
    dtype = utils.infer_dtype_from_bits(n_bits) if pz else None
    effective_numerics = int_numerics.IntNumerics(
        bits=n_bits,
        preserve_zero=pz,
        preserve_max_val=preserve_max_val,
        clip=True,
        round=True,
        noise_fn=None,
        clip_gradient=False,  # This can be disabled when using abs-max scaling.
        dtype=dtype,
    )
  quantizer = Quantizer(
      numerics=effective_numerics,
      calib_shared_axes=None,
      scale_stop_grad=True,
      calibration=calibration.AbsMaxCalibration,
      po2_scale=False,
      context=utils.Context(key=None, train_step=None),
  )
  # TODO(lew): We should try to move to to class constructor or post-init.
  # We currently need to call because bwd pass is too late for initialization.
  if initialize_calibration:
    quantizer.init_calibration()
  return quantizer


def make_fake_quant(quantizer: Quantizer, calibration_axes=None):
  def fake_quant(x):
    x_q, _ = quantizer.quant(x, calibration_axes=calibration_axes)
    return x_q.dequant()

  return fake_quant



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Static range calibration."""

from typing import Sequence

from aqt.jax.v2 import calibration
from aqt.jax.v2 import utils
import flax.linen as nn
from jax import numpy as jnp


_SUM_OF_ONES = "sum_of_ones"
_SUM_OF_VALS = "sum_of_vals"
_MAX_OF_ABS_VALS = "max_of_abs_vals"
_SUM_OF_L1_VALS = "sum_of_l1_vals"
_SUM_OF_LP_VALS = "sum_of_lp_vals"


@utils.flax_slots_kw_only_dataclass
class MeanOfAbsMaxCalibration(calibration.Calibration, nn.Module):
  """State for static range AQT/PTQ/QAT."""

  quant_collection: str

  @nn.compact
  def get_bound(
      self,
      x: jnp.ndarray,
      shared_axes: Sequence[utils.AxisIdx] | None,
      context: utils.Context | None = None,
  ) -> jnp.ndarray:
    abs_max = jnp.max(jnp.abs(x), axis=shared_axes, keepdims=True)
    quant_mode = context.quant_mode if context else utils.QuantMode.TRAIN

    # Initialize the max_stat.value as 0; does not update the max_stat value
    # during initialization.
    is_initializing = not self.has_variable(self.quant_collection, "sum_of_max")

    sum_of_max = self.variable(
        self.quant_collection,
        "sum_of_max",
        jnp.zeros,
        abs_max.shape,
        abs_max.dtype,
    )
    count = self.variable(
        self.quant_collection, "count", jnp.zeros, (), jnp.int32
    )

    if quant_mode == utils.QuantMode.CALIBRATE and not is_initializing:
      sum_of_max.value = sum_of_max.value + abs_max
      count.value = count.value + 1

    # TODO(dhchoi): Introduce checker for count.value.
    # This assertion should be done on jitted graph, and this is not
    # straightforward (or does not have a good option).
    # One option is jax.experimental.checkify
    # (https://jax.readthedocs.io/en/latest/debugging/checkify_guide.html), but
    # we need to transform the whole function with jax.experimental.checkify to
    # use this feature.
    # Maybe wait for the JAX language upgrade to have a better support for this?
    return sum_of_max.value / count.value


# TODO: b/335764538 - Check the math correctness of the module.
class WeightedStatsCalibration(calibration.Calibration, nn.Module):
  """Migration of AQTv1 calibration to AQTv2."""

  quant_collection: str

  # Calibration coefficients
  l1_dev_coeff: float = 0.0
  lp_dev_coeff: float = 0.0
  max_dev_coeff: float = 0.0
  const_bound_coeff: float = 0.0

  # Priors
  update_count_prior: float = 1.0
  max_dev_prior: float = 0.0
  mean_prior: float = 0.0
  l1_dev_prior: float = 0.0
  lp_dev_prior: float = 0.0

  lp_order: int = 2
  safe_divide: bool = False
  filter_zeros: bool = True
  tpu_cross_replica_sum: bool = False
  ema_update_count: float = 100

  def _get_value(self, name: str):
    return self.get_variable(self.quant_collection, name)

  def _mean(self) -> jnp.ndarray:
    sum_of_vals = self._get_value(_SUM_OF_VALS)
    sum_of_ones = self._get_value(_SUM_OF_ONES)
    return self._divide(sum_of_vals, sum_of_ones)

  def _max_dev(self) -> jnp.ndarray:
    return self._get_value(_MAX_OF_ABS_VALS)

  def _l1_dev(self) -> jnp.ndarray:
    sum_of_l1_vals = self._get_value(_SUM_OF_L1_VALS)
    sum_of_ones = self._get_value(_SUM_OF_ONES)
    return self._divide(sum_of_l1_vals, sum_of_ones)

  def _lp_dev(self) -> jnp.ndarray:
    sum_of_lp_vals = self._get_value(_SUM_OF_LP_VALS)
    sum_of_ones = self._get_value(_SUM_OF_ONES)
    if self.lp_order == 2:
      # sqrt() is numerically more accurate
      return jnp.sqrt(self._divide(sum_of_lp_vals, sum_of_ones))
    else:
      # TODO(b/205769820): Make sure if the output of pow op below is
      # numerically valid.
      return self._divide(sum_of_lp_vals, sum_of_ones) ** (1.0 / self.lp_order)

  def _update_var(
      self,
      var: nn.Variable,
      s: jnp.ndarray,
      shared_axes: Sequence[utils.AxisIdx] | None,
      weight: jnp.ndarray | None = None,
      reduce_max: bool = False,
  ):
    """Updates the given Flax variable."""
    if weight is not None:
      if reduce_max:
        # A maximum is a maximum, regardless of its nonzero weight.
        s = jnp.where(weight > 0, s, 0)
      else:
        s = s * weight

    reduce_fn = jnp.max if reduce_max else jnp.sum
    s = reduce_fn(s, axis=shared_axes, keepdims=True)
    if self.tpu_cross_replica_sum:
      raise NotImplementedError(
          "support for tpu_cross_replica_sum=True is not implemented"
      )
    rate = 1.0 / self.ema_update_count
    var.value = (1.0 - rate) * var.value + rate * s

  def _divide(self, x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
    if self.safe_divide:
      # Safe-divide without division-by-zero.
      res = jnp.where(y != 0, x / y, jnp.zeros_like(x))
      return res

    return jnp.divide(x, y)

  @nn.compact
  def get_bound(
      self,
      x: jnp.ndarray,
      shared_axes: Sequence[utils.AxisIdx] | None,
      context: utils.Context | None = None,
  ) -> jnp.ndarray:
    if self.lp_order > 30:
      raise NotImplementedError("For higher norms we should add stabilization.")

    quant_mode = context.quant_mode if context else utils.QuantMode.TRAIN
    stats_shape = list(x.shape)
    if shared_axes is not None:
      for axis in shared_axes:
        stats_shape[axis] = 1

    def _make_var(name: str, init_val: float) -> nn.Variable:
      def init_var_fn(init_val: float) -> jnp.ndarray:
        return jnp.full(stats_shape, init_val, dtype=jnp.float32)

      return self.variable(self.quant_collection, name, init_var_fn, init_val)

    sum_of_ones = _make_var(_SUM_OF_ONES, self.update_count_prior)
    sum_of_vals = _make_var(
        _SUM_OF_VALS, self.mean_prior * self.update_count_prior
    )
    max_of_abs_vals = _make_var(_MAX_OF_ABS_VALS, self.max_dev_prior)
    sum_of_l1_vals = _make_var(
        _SUM_OF_L1_VALS, self.l1_dev_prior * self.update_count_prior
    )
    sum_of_lp_vals = _make_var(
        _SUM_OF_LP_VALS,
        self.lp_dev_prior**self.lp_order * self.update_count_prior,
    )

    if quant_mode in [utils.QuantMode.TRAIN, utils.QuantMode.CALIBRATE]:
      # Update variables.

      # Layers such as ReLU emit zeros often. In such cases, we can model
      # the non-sparse distribution of weights separately, resulting in
      # unbiased estimation of non-sparse mean l1 and lp.
      # This clips away less of the distribution of inputs.
      if self.filter_zeros:
        ones = jnp.where(x != 0, 1.0, 0.0)
      else:
        ones = jnp.ones_like(x)

      px = x if self.lp_order % 2 == 0 else jnp.abs(x)

      self._update_var(sum_of_ones, ones, shared_axes)
      self._update_var(sum_of_vals, x, shared_axes)
      self._update_var(
          max_of_abs_vals, jnp.abs(x), shared_axes, reduce_max=True
      )
      self._update_var(sum_of_l1_vals, jnp.abs(x), shared_axes)
      self._update_var(sum_of_lp_vals, px**self.lp_order, shared_axes)

    return (
        self.l1_dev_coeff * self._l1_dev()
        + self.lp_dev_coeff * self._lp_dev()
        + self.max_dev_coeff * self._max_dev()
        + self.const_bound_coeff
    )
